# $Id: text.rb,v 1.30 2003/07/25 23:41:34 sdalu Exp $

# 
# CONTACT     : zonecheck@nic.fr
# AUTHOR      : Stephane D'Alu <sdalu@nic.fr>
#
# CREATED     : 2002/08/02 13:58:17
# REVISION    : $Revision: 1.30 $ 
# DATE        : $Date: 2003/07/25 23:41:34 $
#
# CONTRIBUTORS: (see also CREDITS file)
#
#
# LICENSE     : GPL v2 (or MIT/X11-like after agreement)
# COPYRIGHT   : AFNIC (c) 2003
#
# This file is part of ZoneCheck.
#
# ZoneCheck is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# ZoneCheck is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ZoneCheck; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

require 'textfmt'
require 'config'

module Publisher
    ##
    ##
    ##
    class Text < Template
	Mime		= "text/plain"
	MaxLineLength	= 79

	##
	## Class for displaying progression information about
	## the tests being performed.
	##
	class Progress
	    class PBar
		BarSize	= 37

		def initialize(output, precision)
		    @output	= output
		    @precision	= precision
		    @mutex	= Mutex::new
		    @updater	= nil
		    @l10n_unit	= $mc.get('pgr_speed_unit')
		end
		
		def start(length)
		    @length     = length
		    @starttime  = @lasttime = Time.now
		    @totaltime  = 0
		    @processed	= 0
		    @tick       = 0

		    @output.write $console.ctl['vi']
		    @output.print barstr

		    @updater	= Thread::new { 
			last_processed = nil
			while true
			    sleep(@precision)
			    nowtime      = Time.now
			    
			    @mutex.synchronize {
				@totaltime      = if @lasttime == @starttime
						  then 0.0
						  else nowtime - @starttime
						  end
				@lasttime       = nowtime

				if @processed != last_processed
				    @tick += 1
				    last_processed = @processed
				end

				@output.write "\r"
				@output.print barstr
				@output.write $console.ctl['ce']
				@output.flush
			    }
			end
		    }
		end
		
		def processed(size)
		    @mutex.synchronize { @processed  += size }
		end

		def done
		    @mutex.synchronize { @updater.kill }
		    @output.write "\r" + $console.ctl['ce']
		    @output.flush
		end
		
		def finish
		    @output.write $console.ctl['ve']
		end
                
		protected
		def barstr
		    speed	= if @totaltime == 0.0
				  then -1.0
				  else @processed / @totaltime
				  end
		    speed_s	= if speed < 0.0 
				  then "--.--%s" % [ @l10n_unit ]
				  else "%7.2f%s" % [ speed, @l10n_unit ]
				  end
		    
		    if @length > 0 then
			pct	= 100 * @processed / @length
			eta	= if speed < 0.0 
				  then -1
				  else ((@length-@processed) / speed).to_i
				  end
			pct_s	= "%2d%%" % pct
			eta_s	= "ETA " + sec_to_timestr(eta)
			bar_s	= '=' * (BarSize * pct / 100) + '>'
			
			"%-4s[%-#{BarSize}.#{BarSize}s] %-11s %10s %12s" % [
			    pct_s, bar_s, @processed, speed_s, eta_s ]
		    else
			ind	= @tick % (BarSize * 2 - 6)
			pos	= if ind < BarSize - 2
				  then ind + 1;
				  else BarSize - (ind - BarSize + 5)
				  end
			bar_s	= " " * BarSize
			bar_s[pos-1,3] = '<=>'
			
			"    [%s] %-11s %10s" % [ bar_s, @processed, speed_s ]
		    end
		end
		
		private
		def sec_to_timestr(sec)
		    return "--:--" if sec < 0
		    
		    hrs = sec / 3600; sec %= 3600;
		    min = sec / 60;   sec %= 60;
		    
		    if (hrs > 0)
		    then sprintf "%2d:%02d:%02d", hrs, min, sec
		    else sprintf "%2d:%02d", min, sec
		    end
		end
	    end


	    # Initialization
	    def initialize(publisher)
		@publisher	= publisher
		@o		= publisher.output
		@counter	= if @publisher.rflag.counter && @o.tty?
				  then PBar::new(@o, 1)
				  else nil
				  end
		@l10n_testing	= $mc.get('w_testing').capitalize
	    end
	    
	    # Start progression
	    def start(count)
		@counter.start(count)	if @counter
	    end
	    
	    # Finished on success
	    def done(desc)
		@counter.done		if @counter
	    end
	    
	    # Finished on failure
	    def failed(desc)
		@counter.done		if @counter
	    end
	    
	    # Finish (finalize) output
	    def finish
		@counter.finish		if @counter
	    end
	    
	    # Process an item
	    def process(desc, ns, ip)
		# Counter
		if @counter
		    @counter.processed(1)
		end

		# Test description
		if @publisher.rflag.testdesc
		    xtra = if    ip then " (IP=#{ip})"
			   elsif ns then " (NS=#{ns})"
			   else          ''
			   end
	    
		    @o.puts "#{@l10n_testing}: #{desc}#{xtra}"
		end
	    end
	end

	#------------------------------------------------------------

	def initialize(rflag, ostream=$stdout)
	    super(rflag, ostream)
	    @count_txt	= $mc.get('title_progress')
	    @progress	= Progress::new(self)
	end


	#------------------------------------------------------------
	
	def error(text)
	    paragraph = ::Text::Format::new
	    paragraph.width = 72
	    paragraph.tag   = $mc.get('w_error').upcase + ': '
	    @o.puts paragraph.format(text)
	end


	def intro(domain)
	    return unless @rflag.intro

	    l10n_zone	= $mc.get('ns_zone').upcase
	    l10n_ns	= $mc.get('ns_ns'  ).upcase
	    sz		= [ l10n_zone.length, l10n_ns.length+3 ].max

	    @o.printf "%-*s : %s\n", sz, l10n_zone, domain.name
	    domain.ns.each_index { |i| 
		n = domain.ns[i]
		@o.printf "%-*s : %s [%s]\n", 
		    sz, i == 0 ? "#{l10n_ns} <=" : "#{l10n_ns}  ",
		    n[0].to_s, n[1].join(', ')
	    }
	    @o.puts
	end

	def diag_start()
	end

	def diag_section(title)
	    txtlen = [title.length, MaxLineLength-20].min
	    txt    = title[0..txtlen]
	    @o.print "       ", "_" * (8+txtlen), "\n"
	    @o.print "     ,", "-" * (8+txtlen), ".|\n"
	    @o.print "~~~~ |    #{txt}    || ", 
		"~" * (MaxLineLength-19-txtlen), "\n"
	    @o.print "     `", "-" * (8+txtlen), "'\n"
	    @o.print "\n"
	end

	def diagnostic1(domainname, 
		i_count, i_unexp, w_count, w_unexp, f_count, f_unexp,
		res, severity)

	    i_tag = @rflag.tagonly ? Config::Info    : $mc.get('w_info_id')
	    w_tag = @rflag.tagonly ? Config::Warning : $mc.get('w_warning_id')
	    f_tag = @rflag.tagonly ? Config::Fatal   : $mc.get('w_fatal_id')
	    
	    i_tag = i_tag.upcase if i_unexp
	    w_tag = w_tag.upcase if w_unexp
	    f_tag = f_tag.upcase if f_unexp

	    summary = "%1s%03d %1s%03d %1s%03d" % [ 
		i_tag, i_count, 
		w_tag, w_count, 
		f_tag, f_count ]

	    printf "%-*s    %s\n", 
		MaxLineLength - 4 - summary.length, domainname, summary

	    if @rflag.tagonly
		@o.puts "  #{severity}: #{res.tag}"
		@o.puts "  #{res.testname}"
	    else
		@o.puts "  #{severity}: #{res.tag}"
		@o.puts "  #{res.desc.msg}"
	    end
	end


	def diagnostic(severity, testname, desc, lst)
	    msg, xpl_lst = nil, nil

	    # Testname
	    if @rflag.testname
		l10n_name = $mc.get("#{testname}_testname")
		@o.puts "[> #{l10n_name}"
	    end

	    # Severity
	    severity_tag		= Config.severity2tag(severity)
	    l10n_severity_shorttag	= if @rflag.tagonly
					  then #{severity_tag}
					  else $mc.get("w_#{severity_tag}_id")
					  end

	    # Message
	    msg = if severity.nil?
		      $mc.get("#{testname}_ok")
		  else
		      if @rflag.tagonly
			  if desc.is_error?
			  then "#{severity}: [Unexpected] #{testname}"
			  else "#{severity}: #{testname}"
			  end
		      else
			  msg = desc.msg
		      end
		  end
	    
	    @o.puts "#{l10n_severity_shorttag}> #{msg}"

	    if !severity.nil?
		# Details
		if @rflag.details && desc.dtl
		    txt = ::Text::Format::new
		    txt.width = 72
		    txt.tag   = "  "
		    txt.format(desc.dtl).split(/\n/).each { |l|
			@o.puts " : #{l}" }
		    @o.puts ' `..... .. .. . .  .'
		end

		# Explanation
		if @rflag.explain && !@rflag.tagonly
		    xpl_lst = xpl_split(desc.xpl)
		end
		
		if xpl_lst
		    xpl_lst.each { |t, h, b|
			tag = $mc.get("tag_#{t}")
			@o.puts " | #{tag}: #{h}"
			b.each { |l| @o.puts " |  #{l}" }
		    }
		    @o.puts ' `----- -- -- - -  -'
		end
	    end
		
	    # Elements
	    lst.each { |elt|
		lines  = elt.split(/\n/)
		
		if !lines.empty?
		    @o.puts "=> #{lines[0]}"
		    lines[1..-1].each { |e|
			@o.puts "   #{e}"
		    }
		end
	    }

	    @o.puts ""
	end
	    

	def status(domainname, i_count, w_count, f_count)
	    @o.printf "==> %s\n", super(domainname, i_count, w_count, f_count)
	end
    end
end
